/*
 * ubtath3kfw.c
 */

/*-
 * Copyright (c) 2010 Maksim Yevmenkin <m_evmenkin@yahoo.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/stdint.h>
#include <sys/stddef.h>
#include <sys/param.h>
#include <sys/queue.h>
#include <sys/types.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/linker_set.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/condvar.h>
#include <sys/sysctl.h>
#include <sys/sx.h>
#include <sys/unistd.h>
#include <sys/callout.h>
#include <sys/malloc.h>
#include <sys/priv.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/firmware.h>

#include "usbdevs.h"
#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usb_ioctl.h>

#define	USB_DEBUG_VAR usb_debug
#include <dev/usb/usb_debug.h>
#include <dev/usb/usb_dev.h>

/*
 * Download firmware to Atheros AR3011.
 */

#define	UBTATH3KFW_FIRMWARE		"ath3k-1.fw"

#define	UBTATH3KFW_IFACE_IDX		0	/* Control interface */
#define	UBTATH3KFW_BSIZE		4096

#define UBTATH3KFW_REQ_DFU_DNLOAD	1

enum {
	UBTATH3KFW_BULK_DT_WR = 0,
	UBTATH3KFW_CTRL_DT_WR,
	UBTATH3KFW_N_TRANSFER,
};

struct ubtath3kfw_softc {
	device_t		sc_dev;
	struct usb_device	*sc_udev;
	struct mtx		sc_mtx;
	struct firmware	const	*sc_fw;
	size_t			sc_fw_offset;
	struct usb_xfer		*sc_xfer[UBTATH3KFW_N_TRANSFER];
};

/*
 * Prototypes
 */

static device_probe_t		ubtath3kfw_probe;
static device_attach_t		ubtath3kfw_attach;
static device_detach_t		ubtath3kfw_detach;

static usb_callback_t		ubtath3kfw_bulk_write_callback;
static usb_callback_t		ubtath3kfw_ctrl_write_callback;

#define ubtath3kfw_xfer_start(sc, i)	usbd_transfer_start((sc)->sc_xfer[(i)])

/*
 * Device's config structure
 */

static const struct usb_config	ubtath3kfw_config[UBTATH3KFW_N_TRANSFER] =
{
	[UBTATH3KFW_BULK_DT_WR] = {
		.type =		UE_BULK,
		.endpoint =	0x02,	/* fixed */
		.direction =	UE_DIR_OUT,
		.if_index =	UBTATH3KFW_IFACE_IDX,
		.bufsize =	UBTATH3KFW_BSIZE,
//		.flags =	{ .pipe_bof = 1, .force_short_xfer = 1, },
		.flags =	{ 0, },
		.callback =	&ubtath3kfw_bulk_write_callback,
		.timeout = 	3000,
	},

	[UBTATH3KFW_CTRL_DT_WR] = {
		.type =		UE_CONTROL,
		.endpoint =	0x00,	/* fixed */
		.direction =	UE_DIR_OUT,
		.if_index =	UBTATH3KFW_IFACE_IDX,
		.bufsize =	sizeof(struct usb_device_request) + 1024,
//		.flags =	{ .pipe_bof = 1, .short_xfer_ok = 1, },
		.flags =	{ 0, },
		.callback =	&ubtath3kfw_ctrl_write_callback,
		.timeout = 	5000,
	},
};

/*
 * Module
 */

static devclass_t	ubtath3kfw_devclass;

static device_method_t	ubtath3kfw_methods[] =
{
	DEVMETHOD(device_probe, ubtath3kfw_probe),
	DEVMETHOD(device_attach, ubtath3kfw_attach),
	DEVMETHOD(device_detach, ubtath3kfw_detach),
	{0, 0}
};

static driver_t		ubtath3kfw_driver =
{
	.name =		"ubtath3kfw",
	.methods =	ubtath3kfw_methods,
	.size =		sizeof(struct ubtath3kfw_softc),
};

DRIVER_MODULE(ubtath3kfw, uhub, ubtath3kfw_driver, ubtath3kfw_devclass, NULL, 0);
MODULE_DEPEND(ubtath3kfw, usb, 1, 1, 1);

/*
 * Probe for a USB Bluetooth device
 */

static int
ubtath3kfw_probe(device_t dev)
{
	const struct usb_device_id	devs[] = {
	/* Atheros AR3011 devices only */
	{ USB_VPI(USB_VENDOR_ATHEROS2, 0x3000, 0) },
	};

	struct usb_attach_arg	*uaa = device_get_ivars(dev);

	if (uaa->usb_mode != USB_MODE_HOST)
		return (ENXIO);

	if (uaa->info.bIfaceIndex != 0)
		return (ENXIO);

	return (usbd_lookup_id_by_uaa(devs, sizeof(devs), uaa));
} /* ubtath3kfw_probe */

/*
 * Attach the device
 */

static int
ubtath3kfw_attach(device_t dev)
{
	struct usb_attach_arg	*uaa = device_get_ivars(dev);
	struct ubtath3kfw_softc	*sc = device_get_softc(dev);
	uint8_t			iface_index;
	int			error;

	sc->sc_dev = dev;
	sc->sc_udev = uaa->device;

	device_set_usb_desc(dev);

	mtx_init(&sc->sc_mtx, "ubtath3kfw lock", NULL, MTX_DEF | MTX_RECURSE);
	sc->sc_fw = NULL;
	sc->sc_fw_offset = 0;

	iface_index = UBTATH3KFW_IFACE_IDX;
	error = usbd_transfer_setup(uaa->device, &iface_index, sc->sc_xfer,
				ubtath3kfw_config, UBTATH3KFW_N_TRANSFER,
				sc, &sc->sc_mtx);
	if (error != 0) {
		device_printf(dev, "allocating USB transfers failed. %s\n",
			usbd_errstr(error));
		mtx_destroy(&sc->sc_mtx);
		return (ENXIO);
	}

	sc->sc_fw = firmware_get(UBTATH3KFW_FIRMWARE);
	if (sc->sc_fw == NULL) {
		device_printf(dev, "could not locate firmware image %s\n",
			UBTATH3KFW_FIRMWARE);
		ubtath3kfw_detach(dev);
		return (ENXIO);
	}

	mtx_lock(&sc->sc_mtx);
	ubtath3kfw_xfer_start(sc, UBTATH3KFW_CTRL_DT_WR);
	mtx_unlock(&sc->sc_mtx);

	return (0);	/* success */
} /* ubtath3kfw_attach */ 

/*
 * Detach the device
 */

static int
ubtath3kfw_detach(device_t dev)
{
	struct ubtath3kfw_softc	*sc = device_get_softc(dev);

	if (sc->sc_fw != NULL) {
		firmware_put(sc->sc_fw, FIRMWARE_UNLOAD);
		sc->sc_fw = NULL;
	}

	usbd_transfer_unsetup(sc->sc_xfer, UBTATH3KFW_N_TRANSFER);
	mtx_destroy(&sc->sc_mtx);

	return (0);
} /* ubtath3kfw_detach */

/*
 * USB write callback (bulk)
 */

static void
ubtath3kfw_bulk_write_callback(struct usb_xfer *xfer, usb_error_t error)
{
	struct ubtath3kfw_softc	*sc = usbd_xfer_softc(xfer);
	struct usb_page_cache	*pc;
	int			actlen;
	size_t			togo;

	switch (USB_GET_STATE(xfer)) {
	case USB_ST_TRANSFERRED:
		usbd_xfer_status(xfer, &actlen, NULL, NULL, NULL);
		sc->sc_fw_offset += actlen;
		/* FALLTHROUGH */

	case USB_ST_SETUP:
setup_next:
		togo = sc->sc_fw->datasize - sc->sc_fw_offset;
		if (togo == 0)
			break;
		if (togo > UBTATH3KFW_BSIZE)
			togo = UBTATH3KFW_BSIZE;

		pc = usbd_xfer_get_frame(xfer, 0);
		usbd_copy_in(pc, 0,
			(const unsigned char *) sc->sc_fw->data +
				sc->sc_fw_offset, togo);
		usbd_xfer_set_frame_len(xfer, 0, togo);

		usbd_transfer_submit(xfer);
		break;

	default: /* Error */
		if (error != USB_ERR_CANCELLED) {
			usbd_xfer_set_stall(xfer);
			goto setup_next;
		}
		break;
	}
} /* ubtath3kfw_bulk_write_callback */

/*
 * USB write callback (control)
 */

static void
ubtath3kfw_ctrl_write_callback(struct usb_xfer *xfer, usb_error_t error)
{
	struct ubtath3kfw_softc		*sc = usbd_xfer_softc(xfer);
	struct usb_device_request	req;
	struct usb_page_cache		*pc;
	int				actlen;

	switch (USB_GET_STATE(xfer)) {
	case USB_ST_SETUP:
		bzero(&req, sizeof(req));
		req.bmRequestType = UT_VENDOR;
		req.bRequest = UBTATH3KFW_REQ_DFU_DNLOAD;
		USETW(req.wLength, 20);

		pc = usbd_xfer_get_frame(xfer, 0);
		usbd_copy_in(pc, 0, &req, sizeof(req));
		usbd_xfer_set_frame_len(xfer, 0, sizeof(req));

		pc = usbd_xfer_get_frame(xfer, 1);
		usbd_copy_in(pc, 0, sc->sc_fw->data, 20);
		usbd_xfer_set_frame_len(xfer, 1, 20);

		usbd_xfer_set_frames(xfer, 2);

		usbd_transfer_submit(xfer);
		break;

	case USB_ST_TRANSFERRED:
		usbd_xfer_status(xfer, &actlen, NULL, NULL, NULL);
		sc->sc_fw_offset += (actlen - sizeof(req));

		ubtath3kfw_xfer_start(sc, UBTATH3KFW_BULK_DT_WR);
		break;

	default: /* Error */
		if (error != USB_ERR_CANCELLED)
			usbd_xfer_set_stall(xfer);
		break;
	}
} /* ubtath3kfw_ctrl_write_callback */
